// Copyright Marcus Del Favero 2025
// Licensed under the GNU AGPLv3 with an exception, see `README.md` for details
use std::{mem, ops::AddAssign};

use glam::{UVec2, IVec2, Vec4};

use super::image::Image;

#[derive(Clone)]
pub struct Layer<T> {
	size: UVec2,
	data: Box<[T]>,
}

pub struct Kernel(Box<[f32]>);

impl<T: Default + Clone> Layer<T> {
	pub fn clear(&mut self) {
		self.data.fill(T::default());
	}
}

impl<T: Copy> Layer<T> {
	/**
	 * Gets the pixel at the provided position, which is clamped if out of
	 * bounds.
	 */
	pub fn get(&self, x: i32, y: i32) -> T {
		let pos = IVec2::new(x.clamp(0, self.size.x as i32 - 1), y.clamp(0, self.size.y as i32 - 1)).as_uvec2();
		self.data[(pos.x + pos.y * self.size.x) as usize]
	}
}

impl<T> Layer<T> {
	/**
	 * Sets the pixel at the provided position, panicking of out of bounds.
	 */
	pub fn set(&mut self, x: u32, y: u32, val: T) {
		assert!(x < self.size.x && y < self.size.y);
		self.data[(x + y * self.size.x) as usize] = val;
	}
}

impl<T: AddAssign<T>> Layer<T> {
	pub fn point(&mut self, x: u32, y: u32, amount: T) {
		assert!(x < self.size.x && y < self.size.y);
		self.data[(x + y * self.size.x) as usize] += amount;
	}
}

impl Layer<Vec4> {
	pub fn new(width: u32, height: u32, colour: Vec4) -> Layer<Vec4> {
		Layer {
			size: UVec2::new(width, height),
			data: vec![colour; (width * height) as usize].into_boxed_slice(),
		}
	}

	/**
	 * Replaces all RGB values with black, keeping the alpha value.
	 */
	pub fn blacken(&mut self) {
		for pixel in &mut self.data {
			for i in 0..3 {
				pixel[i] = 0.0;
			}
		}
	}

	pub fn gaussian_blur_kernel(stddev: f32) -> Kernel {
		assert!((0.0..1000.0).contains(&stddev)); // Shouldn't be too big

		let half = (stddev * 3.0).ceil() as usize;
		let size = half * 2 + 1;
		let mut data = Vec::with_capacity(size);

		for i in 0..half {
			let x = (half - i) as f32;
			data.push((-x.powi(2) / (2.0 * stddev.powi(2))).exp());
		}

		let scale = 1.0 / (data.iter().sum::<f32>() * 2.0 + 1.0);
		for v in &mut data { *v *= scale; }

		data.push(scale); // Adds the centre
		for i in (0..half).rev() { data.push(data[i]); }

		let kernel = data.into_boxed_slice();
		assert!(kernel.len() % 2 == 1);
		assert!((kernel.iter().sum::<f32>() - 1.0).abs() < 1e-4);
		Kernel(kernel)
	}

	pub fn gaussian_blur(&mut self, kernel: &Kernel) {
		let mut old = self.clone();

		let off = -(kernel.0.len() as i32 / 2);

		// Does this on each axis using a 1D kernel
		// First does the x-axis, then the y-axis
		for y in 0..self.size.y {
			for x in 0..self.size.x {
				let mut val = Vec4::ZERO;
				for (i, &k) in kernel.0.iter().enumerate() {
					val += old.get(x as i32 + off + i as i32, y as i32) * k;
				}
				self.set(x, y, val);
			}
		}

		mem::swap(&mut old, self);

		for x in 0..self.size.x {
			for y in 0..self.size.y {
				let mut val = Vec4::ZERO;
				for (i, &k) in kernel.0.iter().enumerate() {
					val += old.get(x as i32, y as i32 + off + i as i32) * k;
				}
				self.set(x, y, val);
			}
		}
	}

	/**
	 * Adds the layer `other` on top of this layer, with the corner of `layer`
	 * being positioned at `offset` in `self`.
	 */
	pub fn add(&mut self, other: &Layer<Vec4>, offset: IVec2) {
		for y in offset.y..offset.y + other.size.y as i32 {
			if y >= self.size.y as i32 { break; }
			for x in offset.x..offset.x + other.size.x as i32 {
				if x >= self.size.x as i32 { break; }
				let (src, dst) = (other.get(x - offset.x, y - offset.y), self.get(x, y));
				let mut colour = Vec4::ZERO;

				// https://en.wikipedia.org/wiki/Alpha_compositing#Description
				colour.w = src.w + dst.w * (1.0 - src.w);
				for i in 0..3 { colour[i] = (src[i] * src.w + dst[i] * dst.w * (1.0 - src.w)) / colour.w; }
				self.set(x as u32, y as u32, colour);
			}
		}
	}
}

impl Layer<f32> {
	pub fn new(width: u32, height: u32) -> Layer<f32> {
		Layer {
			size: UVec2::new(width, height),
			data: vec![0.0; (width * height) as usize].into_boxed_slice(),
		}
	}

	pub fn normalise(&mut self) {
		let max = *self.data.iter().max_by(|a, b| a.total_cmp(b)).unwrap();
		for v in &mut self.data { *v /= max; }
	}

	pub fn into_pixels(self) -> Vec<f32> {
		self.data.into_vec()
	}

	pub fn darken(&mut self) {
		let fsize = self.size.as_vec2();
		let centre = fsize / 2.0 - 0.5;

		let mut i = 0;
		for y in 0..self.size.y {
			for x in 0..self.size.x {
				let dist = ((UVec2::new(x, y).as_vec2() - centre) / (fsize / 2.0)).length();
				self.data[i] *= if dist >= 1.0 {
					0.0
				} else {
					1.0 - dist * dist
				};
				i += 1;
			}
		}
	}
}

impl From<Image> for Layer<Vec4> {
	fn from(image: Image) -> Layer<Vec4> {
		assert_eq!(image.data.len(), (image.width * image.height * 4) as usize, "Expected RGBA image");

		let mut data = Vec::with_capacity((image.width * image.height) as usize);
		for pixel in image.data.chunks_exact(4) {
			let mut c = Vec4::ZERO;
			for i in 0..4 { c[i] = pixel[i] as f32 / 255.0; }
			data.push(c);
		}

		Layer {
			size: UVec2::new(image.width, image.height),
			data: data.into_boxed_slice(),
		}
	}
}

impl From<Layer<Vec4>> for Image {
	fn from(layer: Layer<Vec4>) -> Image {
		let mut data = Vec::with_capacity((layer.size.element_product() * 4) as usize);
		for pixel in &layer.data {
			for i in 0..4 { data.push(u8::try_from((pixel[i] * 255.0).round() as i32).expect("Layer values must be between 0 and 1")); }
		}

		Image {
			width: layer.size.x,
			height: layer.size.y,
			data: data.into_boxed_slice(),
		}
	}
}
