// Copyright Marcus Del Favero 2025
// Licensed under the GNU AGPLv3 with an exception, see `README.md` for details
use png::ColorType;

use super::image::Image;
use super::inkscape;

use crate::utils;

const INPUT: &str = "data/player.svg";
const OUTPUT: &str = "../data/textures/player.png";

const BASE_PATH: &str = "/tmp/base.png";
const COLOUR_PATH: &str = "/tmp/colour.png";
const ALPHA_PATH: &str = "/tmp/alpha.png";

pub fn generate() {
	const SIZE: u32 = 256;

	if utils::newer(INPUT, OUTPUT) {
		let output = generate_image(SIZE);
		output.save(OUTPUT, ColorType::GrayscaleAlpha);
	}
}

pub fn generate_icon(size: u32, path: &str) {
	let input = generate_image(size);

	inkscape::gen_layer_grey(INPUT, ALPHA_PATH, "alpha", size, size / 2);
	let alpha_input = Image::load(ALPHA_PATH, ColorType::Grayscale);

	let mut output = Image::new(size, size, 4);
	let player_colour = [255, 63, 31];

	assert!(input.width == size && input.height == size / 2);

	for y in 0..size / 2 {
		for x in 0..size {
			let i = (x + y * size) as usize;
			let base = input.data[i * 2];
			let colour = input.data[i * 2 + 1];
			let alpha = alpha_input.data[i];
			let t = colour as f64 / 255.0;

			let out_indices = [
				(x + y * size) as usize,
				(x + (size - 1 - y) * size) as usize,
			];

			for out_i in out_indices {
				for (i, &colour) in player_colour.iter().enumerate() {
					output.data[out_i * 4 + i] = (base as i32 + ((colour - base as i32) as f64 * t) as i32) as u8;
				}

				output.data[out_i * 4 + 3] = alpha;
			}
		}
	}

	output.save(path, ColorType::Rgba);
}

fn generate_image(size: u32) -> Image {
	inkscape::gen_layer_grey(INPUT, BASE_PATH, "base", size, size / 2);
	inkscape::gen_layer_grey(INPUT, COLOUR_PATH, "colour", size, size / 2);

	let (width, height, base, colour) = {
		let base = Image::load(BASE_PATH, ColorType::Grayscale);
		let colour = Image::load(COLOUR_PATH, ColorType::Grayscale);

		assert!(base.width == colour.width && base.height == colour.height, "Base and colour layers aren't the same dimensions");
		(base.width, base.height, base.data, colour.data)
	};

	let mut output = Image::new(width, height, 2);
	for i in 0..(width * height) as usize {
		(output.data[i * 2], output.data[i * 2 + 1]) = (base[i], colour[i]);
	}
	output
}
